const Factory = require('./plugin/factory');
const Logger = require('./log');
const Config = require('./config');
const Shell = require('./shell');
const Prompt = require('./prompt');
const Spinner = require('./spinner');
const Metrics = require('./metrics');
const { reduceUntil, parseVersion } = require('./util');
const handleDeprecated = require('./deprecated');

const runTasks = async (options, di) => {
  const container = Object.assign({}, di);
  container.config = container.config || new Config(options);

  const { config } = container;
  const { isInteractive, isVerbose, isDryRun, isDebug } = config;
  const global = { isInteractive, isVerbose, isDryRun, isDebug };

  container.log = container.log || new Logger(global);
  container.spinner = container.spinner || new Spinner({ global, container, config });
  container.prompt = new Prompt({ container: { config } });
  container.metrics = new Metrics({ isEnabled: config.isCollectMetrics });
  container.shell = container.shell || new Shell({ global, container });

  try {
    const { log, metrics, shell, spinner, prompt } = container;

    const options = handleDeprecated(config.getContext(), log);
    metrics.trackEvent('start', options);

    const plugins = await Factory.getPlugins(config, global, { config, log, shell, spinner, prompt });

    await Promise.all(plugins.map(plugin => plugin.init()));

    const { beforeStart, beforeBump, afterBump, beforeStage, afterRelease } = options.scripts;

    const name = await reduceUntil(plugins, plugin => plugin.getName());
    const latestVersion = await reduceUntil(plugins, plugin => plugin.getLatestVersion());

    // TODO: what's the best way to get (and store) this information?
    const latestTag = await reduceUntil(plugins, plugin => plugin.getContext('latestTagName'));
    const repo = await reduceUntil(plugins, plugin => plugin.getContext('repo'));
    config.setContext({ name, latestVersion, latestTag, repo });

    // Helper for user scripts
    const runScript = script => {
      if (!script) return;
      const context = config.getContext();
      const task = () => shell.exec(script, {}, context);
      return spinner.show({ task, label: script, context, forced: true });
    };

    await runScript(beforeStart);

    log.obtrusive(`🚀 Let's release ${name} (currently at ${latestVersion})`);

    const version = await reduceUntil(plugins, async plugin => plugin.getIncrementedVersion({ latestVersion }));
    config.setContext(parseVersion(version));

    for (const plugin of plugins) {
      await plugin.beforeBump();
    }

    // TODO: what's the best way to get (and store) this information?
    const changelog = await reduceUntil(plugins, plugin => plugin.getContext('changelog'));
    config.setContext({ changelog });

    await runScript(beforeBump);
    await Promise.all(plugins.map(plugin => plugin.bump(version)));
    await runScript(afterBump);
    await runScript(beforeStage);

    for (const plugin of plugins) {
      await plugin.beforeRelease();
    }

    plugins.reverse();

    for (const plugin of plugins) {
      await plugin.release();
    }

    await runScript(afterRelease);

    for (const plugin of plugins) {
      await plugin.afterRelease();
    }

    await metrics.trackEvent('end');

    log.log(`🏁 Done (in ${Math.floor(process.uptime())}s.)`);

    return {
      name,
      changelog,
      latestVersion,
      version
    };
  } catch (err) {
    const { log, metrics } = container;
    await metrics.trackException(err);
    log.error(err.message || err);
    throw err;
  }
};

module.exports = runTasks;
